package OME;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.lang.*;
import java.util.*;
import java.io.*;

/** This class provides an interface to display and edit a graphic view, 
 *  as well as access other relevant functionalities. 
 *  It then interacts directly with <code> GraphicView </code>, which is a class 
 *  storing all graphic view related information, to realize user's requests.
 * 
 *  @author Ying Shi
 *  @version 28 Aug 1998
 *
 *  @see GraphicView
 *  @see ScrollingPanel
 */
class GraphicViewCanvas extends JPanel {

    static final float ICON_HEIGHT = (float)16.0;
    
    private JPopupMenu popupmenu;
    private Map index2type;
    
    private JMenuBar menuBar;
    private JToolBar toolBar;
    private GraphicView view;
    private GraphicViewFrame gframe;
    private OME ome;
    
    private String m_sCurrFile = "";
    private String m_sCurrDir = "";
    
    private Point clickedp;	    // Used only for mouse drags.

    // The types of OMEFramework.Method.Paramter we support.
    static Class stringClass;
    static Class integerClass;
    static Class LocationClass;
    static Class GraphicLocationClass;
    static Class OMEObjectClass;
    static Class OMEElementClass;
    static Class OMELinkClass;
    static Class OMEContextClass;

    private Collection plugins;

    // MODES
    private static final int NORMAL_MODE = 0;
    private static final int GET_PARAMETER = 1;
    private int mode;
    private GVCParameterCollector pcollector;
    
    /** Takes care of parameter collecting for methods.*/
    private class GVCParameterCollector {
    
	GraphicViewCanvas gvc;
	private OMEMethod method = null;
	private Class classType;
	private Vector v;
	private JToggleButton button = null;
	private OMEContext context;
	
	public GVCParameterCollector(GraphicViewCanvas gvc) {
	    this.gvc = gvc;
	}

	/** Initiates parameter collecting.
	 *  @param method   The method for whom we are collecting parameters.
	 */
	public void startCollecting(OMEMethod method, OMEContext context,
		JToggleButton button) {
	    if (method == null) {
		D.a("Can't collect for a null method.");
	    }
	    
	    if (this.method != null) {
		// User has started a method while in the midst of another.
		// Cancel the one we were working on.
		cancelMethod();

		// Now we can start the new one.
	    }
	    
	    this.method = method;
	    this.context = context;
	    v = new Vector();

	    OMEParameter p = method.nextParameter();
	    if (p == null) {
		method.invoke();
	    } else {
		getParameter(p);
	    }

	    this.button = button;
	}
	
	/** Prepare the application to acquire the desired parameter. */
	private void getParameter(OMEParameter p) {
	    classType = p.getType();
	    
	    if (classType == stringClass) {
		// Ignore cardinality (for now)
		String s = JOptionPane.showInputDialog(gvc, p.getMessage());
		v.add(s);
		reportParameter();
	    } else if (classType == integerClass) {
		String s;
		Integer i = null;
		while (i == null) {
		    try {
			s = JOptionPane.showInputDialog(gvc, p.getMessage());
			i = new Integer(s);
		    } catch (Exception e) {
			// User did not enter an integer.  Try again.
			// A beep might be nice.
		    }
		}
		v.add(i);
		reportParameter();
	    } else if (classType == OMEContextClass) {
		D.o("Feeding a context.");
		//OMEContext con = view.generateContext(getRightClickedP());
		v.add(context);
		reportParameter();
	    } else {
		//enter GET_PARAMETER mode
		gvc.setMode(GET_PARAMETER, p.getMessage());
	    }
	}

	/** Feeds the parameter we've been collecting to our method, and
	 *  handles the next parameter asked for by our method. */
	private void reportParameter() {
	    method.passParameter(v);
	    OMEParameter p = method.nextParameter();

	    // We're done with v now, kill it. 
	    v = new Vector();
	    
	    if (p == null) {
		//We're done here, invoke the method
		method.invoke();
		resetPC();
	    } else {
		//Keep on truckin'.
		getParameter(p);
	    }
	}

	/** Receive mousepressed events from the user (while in GET_PARAMETER
	 *  mode).
	 */
	public void processMousePressed(MouseEvent me) {
	    Point p = view.toView(me.getPoint());
	    boolean foundone = false;
	    
	    if (classType==LocationClass || classType==GraphicLocationClass) {
		View.GraphicLocation gl = new View.GraphicLocation();
		gl.setPoint(p);
		v.add(gl);
		foundone = true;
	    } else if (classType == OMEObjectClass) {
		GraphicViewObject gvo = view.findObject(p);
		if (gvo != null) {
		    v.add(gvo);
		    foundone = true;
		}
	    } else if (classType == OMEElementClass) {
		GraphicViewElement gve = view.findElement(p);
		if (gve != null) {
		    v.add(gve);
		    foundone = true;
		}
	    } else if (classType == OMELinkClass) {
		GraphicViewLink gvl = view.findLink(p);
		if (gvl != null) {
		    v.add(gvl);
		    foundone = true;
		}
	    } else {
		// we don't support this type of parameter.
	    }
	    if (foundone) {
		// Do we stay in this mode or return to normal mode?
		// We should probably look at the CTRL_MASK etc., but at
		// the moment we simply go back to normal mode.
		gvc.setMode(NORMAL_MODE,null);

		// Again with the ignoring cardinality...
		reportParameter();
	    }
	}

	/** The user has abandoned the method he was in the middle of.  Abort
	 *  abort!
	 */
	public void cancelMethod() {
	    gvc.setMode(NORMAL_MODE,null);
	    method.cancelled();
	    resetPC();
	}

	/** Resets the parameter collector so that it is ready to start a new
	 *  method.
	 */
	private void resetPC() {
	    //Untoggle our button if we had one.
	    if (button != null) {
		button.setSelected(false);
	    }
	    // Clear our fields. 
	    method = null;
	    button = null;
	}
	
    }
    
    /** Constructs the canvas.
     *
     *  @param menuBar 	the menu bar created by <code> GraphicViewFrame
     *			    </code>.
     *  @param toolBar 	the tool bar created by <code> GraphicViewFrame
     *			    </code>.
     *  @param gview 	the <code> GraphicView </code> object. It better not
     *			be null.
     *  @param ome 	the main window.
     *  @param gframe 	the main graphic view frame window.
     */
    public GraphicViewCanvas(JMenuBar menuBar, GraphicView gview, OME ome,
	GraphicViewFrame gframe) {

	try {
	    stringClass = Class.forName("java.lang.String");
	    integerClass = Class.forName("java.lang.Integer");
	    LocationClass = Class.forName("OME.View$Location");
	    GraphicLocationClass = Class.forName("OME.View$GraphicLocation");
	    OMEObjectClass = Class.forName("OME.OMEObject");
	    OMEElementClass = Class.forName("OME.OMEElement");
	    OMELinkClass = Class.forName("OME.OMELink");
	    OMEContextClass = Class.forName("OME.OMEContext");
	} catch (Exception e) {
	    D.o(e);
	}
	
	D.o("Creating GraphicViewCanvas");

	this.gframe = gframe;
	this.ome = ome;
	this.menuBar = menuBar;
	//this.toolBar = toolBar;
	isTabbed = false;
	view = gview;
	pcollector = new GVCParameterCollector(this);
	
	//Load the plugin.  This is not the right place to do it.
	plugins = new Vector();
	plugins.add(new OMEDefaultPlugin(view.getModel()));
	plugins.add(new IStarPlugin(view.getModel()));
	
	index2type = new HashMap();
	popupmenu = null; // setPopupMenu();
	setMenuBar();
	setToolBar();

	// setup the GraphicView.
	view.setDisplayBounds(new Rectangle(new Point(0,0),gframe.getSize()));
	view.setGraphicViewCanvas(this);
    
	// Gui stui.
	setLayout(null);
	setBorder(BorderFactory.createEtchedBorder());
	addComponentListener( new GVCComponentAdapter() );
	addMouseListener( new GVCMouseAdapter() );
	addMouseMotionListener( new GVCMouseMotionAdapter());
	addKeyListener( new GVCKeyAdapter());
	requestFocus();
	
	D.o("Created GraphicViewCanvas");
    }

    /** Recursively populates the given Menu using the OMEPlugin/Method
     *  interface. 
     *
     *  @param menu	    The menu we are populating.
     *  @param menumethod   The OMEMethod we are using to populate the menu.
     *  @param p	    The point associated with initiations of methods
     *			    from this menu (used by makeMenuItem).
     */
    private void populateMenu(JMenu menu, OMEMethod menumethod, Point p) {
	D.o("Populating Menu");
	Collection c = menumethod.getSubmenu();
	if (c != null) {
	    Iterator i = c.iterator();
	    while (i.hasNext()) {
		OMEMethod method = (OMEMethod)i.next();
		//method may be a normal method, or a submenu.
		Collection subcollection = method.getSubmenu();
		if (subcollection != null) {
		    JMenu submenu = new JMenu(method.getName());
		    populateMenu(submenu, method, p);
		} else {
		    // Normal case.  Hook up the button.
		    JMenuItem jmi = makeMenuItem(method, p);
		    menu.add(jmi);
		}
	    }
	}
    }

    /** Returns a JMenuItem with the given OMEMethod "hooked up".
     *  @param method	A "non-submenu" OMEMethod.
     *  @param p	The point associated with the initiation of this
     *                  method (used by the ActionAdapter).
     */
    private JMenuItem makeMenuItem(OMEMethod method, Point p) {
	JMenuItem jmi = new JMenuItem(method.getName());
	Image im = method.getImage();
	if (im != null) {
	    ImageIcon ic = iconifyImage(im);
	    jmi.setIcon(ic);
	}
	jmi.addActionListener(new GVCActionAdapter(method, p, this));
	return jmi;
    }

    /** Scales image and converts ito an ImageIcon. */
    private ImageIcon iconifyImage(Image im) {
	ImageTable it = view.getImageTable();
	im = it.getScaling(im,ICON_HEIGHT/it.getHeight(im));
	return new ImageIcon(im);
    }

    /** Creates and populates the popup menu.
     *  @return 	the newly created popup menu.
     */
    protected JPopupMenu setPopupMenu(Point p) {
	D.o("Creating popup");
	
	JPopupMenu popupmenu = new JPopupMenu();
	
	Iterator plugit = plugins.iterator();
	Collection c = new Vector();
	while (plugit.hasNext()) {
	    Collection methods = 
		    ((OMEPlugin)plugit.next()).getPopupMethods(view);
	    if (methods != null) {
		c.addAll(methods);
	    }
	}
	
	// Too bad JPopupMenus aren't JMenus.
	Iterator i = c.iterator();
	while (i.hasNext()) {
	    OMEMethod method = (OMEMethod)i.next();
	    //method may be a normal method, or a submenu.
	    Collection subcollection = method.getSubmenu();
	    if (subcollection != null) {
		JMenu submenu = new JMenu(method.getName());
		populateMenu(submenu, method, p);
		popupmenu.add(submenu);
	    } else {
		// Normal case.  Hook up the button.
		JMenuItem jmi = makeMenuItem(method, p);
		popupmenu.add(jmi);
		OMEContext con = view.generateContext(p);
		jmi.setEnabled(method.isEnabled(con));
	    }
	}
	return popupmenu;
    }
    

    /** Populates the menu bar created by and passed from <code>
     * GraphicViewFrame</code>.
     */
    protected void setMenuBar() {

	Iterator plugit = plugins.iterator();
	Collection c = new Vector();
	while (plugit.hasNext()) {
	    Collection methods = 
		    ((OMEPlugin)plugit.next()).getMenubarMethods(view);
	    if (methods != null) {
		c.addAll(methods);
	    }
	}
	Iterator i = c.iterator();
	while (i.hasNext()) {
	    OMEMethod menumethod = (OMEMethod)i.next();
	    // We assume that all methods are menuMethods.
	    JMenu menu = new JMenu(menumethod.getName());
	    populateMenu(menu, menumethod, null);
	    menuBar.add(menu);
	}
    }    


    /** Populates the tool bar, which is created by and passed from <code>
     *  GraphicViewFrame </code>.
     */
    protected void setToolBar() {
	
	Iterator plugit = plugins.iterator();
	Collection c = new Vector();
	while (plugit.hasNext()) {
	    Collection methods = 
		    ((OMEPlugin)plugit.next()).getToolbarMethods(view);
	    if (methods != null) {
		c.addAll(methods);
	    }
	}
	Iterator i = c.iterator();

	while (i.hasNext()) {
	    //We're going to add a new toolbar...
	    OMEMethod barmethod = (OMEMethod)i.next();
	    JToolBar newbar = new JToolBar();
	    newbar.setName(barmethod.getName());
	    Iterator ii = barmethod.getSubmenu().iterator();
	    while (ii.hasNext()) {
		OMEMethod method = (OMEMethod)ii.next();
		// Do Image Stuff
		Image im = method.getImage();
		ImageIcon ic = iconifyImage(im);
		// Add button
		JToggleButton jb = new JToggleButton(ic);
		jb.setRequestFocusEnabled(false);
		jb.setMargin(new Insets(1,1,1,1));
		jb.setToolTipText(method.getName());
		jb.addActionListener(new GVCActionAdapter(method, jb, this));
		newbar.add(jb);
		//toolBar.add(jb); // everything goes on the same bar, for now.
	    }

	    gframe.addToolBar(newbar);
	    //toolBar.addSeparator();
	}
}

    /** Returns the private member, view.
     *  @return Private member view, the <code> GraphicView </code> object.
     */
    public GraphicView getData() {
	return view;
    }
 
    // Vertical and horizontal scrollbars
    //
    Scrollbar sbv; 
    Scrollbar sbh;
    

    /** Sets the view value in <code> GraphicView </code> according to the value of the
     *  vertical scroll bar.
     *  @param value 	the value of the vertical scroll bar.
     */
    public void setSbvValue(int value) {

	view.getDisplayBounds().y = value;//+view.getViewBounds().y;
	//adjustScrollBar();
	view.refreshVisible();
    }


    /** Sets the view value in <code> GraphicView </code> according to the value of the
     *  horizontal scroll bar.
     *  @param value 	the value of the horizontal scroll bar.
     */
    public void setSbhValue(int value) {

	view.getDisplayBounds().x = value;//+view.getViewBounds().x;
	//adjustScrollBar();
	view.refreshVisible();
    }


    /** Sets the value of the scroll bars according to the view in <code>
     *  GraphicView </code>.
     */
    public void adjustScrollBar() {
	sbh.setValues( view.getDisplayBounds().x, 
		view.getDisplayBounds().width, view.getViewBounds().x,
		view.getViewBounds().x + view.getViewBounds().width);
	sbh.setBlockIncrement( view.getDisplayBounds().width/2 );
	
	sbv.setValues( view.getDisplayBounds().y,
		view.getDisplayBounds().height, view.getViewBounds().y,
	       	view.getViewBounds().y + view.getViewBounds().height);
	sbv.setBlockIncrement( view.getDisplayBounds().height/2 );
    }

    
    /** Passes the scroll bar objects to this class.
     *
     *  @param psbv  the vertical scroll bar object.
     *  @param psbh  the horizonal scroll bar object.
     */
    public void SetScrollbars( Scrollbar psbv, Scrollbar psbh ) {
	sbv = psbv;
	sbh = psbh;
    }

    private int pressed; // die die die!

    /** Toggles selection.
     *  @param p 	the last clicked point.
     */
    public void toggleSelect(Point p) {
	D.o("ToggleSelecting");
	GraphicViewElement ge;
	GraphicViewLink gl;

	setClickedP(p);
	
	Iterator i = view.getSelectedElements().iterator();
	while (i.hasNext()) {
	    GraphicViewElement gve = (GraphicViewElement)i.next();
	    D.o("Selected Els include: "+gve.getName());
	}
	
	
	if((ge=view.findElement(p))!=null) {
	    if(view.getSelectedElements().contains(ge)) {
		    view.getSelectedElements().remove(ge);
		    D.o("Deselected: "+ge.getName());
	    } else {
		view.getSelectedElements().add(ge);
		D.o("Selecting: "+ge.getName());
		
	    }
	}

	if((gl=view.findLink(p)) != null) {
	    if(view.getSelectedLinks().contains(gl)) {
		    view.getSelectedLinks().remove(gl);
		    D.o("Deselecting: " + gl.getName());
	    } else {
		view.getSelectedLinks().add(gl);
		D.o("Selecting: "+gl.getName());
	    }
	}
    }


    /** Makes selection.
     *  @param p 	the last clicked point.
     */
    public void select (Point p) {
	GraphicViewElement ge;
	GraphicViewLink gl;
	setClickedP(p);
	
	view.deselectAll();
	D.o("Deselecting all.");
	
	if((ge=view.findElement(p)) != null) {
	    view.getSelectedElements().add(ge);
	} else if((gl=view.findLink(p)) != null) {
	    view.getSelectedLinks().add(gl);
	}
    }


    /** Draws an expanded element.
     *  @param g	the <code> Graphics </code> to draw on.
     *  @param ge	the expanded element to be drawn.
     */
    public void drawExpandedElements(Graphics2D g, GraphicViewElement ge) {
	g.setColor(Color.lightGray);
	
	Rectangle rec = ge.getExpandedBounds();
	g.fillArc(rec.x+2, rec.y+2, rec.width-4, rec.height-4, 0, 360);
	g.setColor(Color.black);
	g.drawArc(rec.x+2, rec.y+2, rec.width-4, rec.height-4, 0, 360);
    
	// Draws the element image.
	drawElements(g, ge); 
    }
    

    /** Draws an element.
     *  @param g 	the <code> Graphics </code> to draw on.
     *  @param ge	the element to be drawn.
     */
    public void drawElements(Graphics g, GraphicViewElement ge) {
	//clear the area (use GVObject.biggestBounds(?))
//	g.clearRect(ge.getBounds().x,ge.getBounds().y, 
//		(int)ge.getBounds().getWidth(), (int)ge.getBounds().getHeight());
	g.drawImage(ge.getImage(), ge.getImageBounds().x,
		ge.getImageBounds().y, this);
	
	// Draws the name of the element.
		
	// We really should ask the element for its font.
	FontMetrics fm = g.getFontMetrics();
	int fh = fm.getAscent() + fm.getDescent();
	
	Collection c = ge.getText();
	if (c != null) {
	    Iterator i = c.iterator();
	    int y = (ge.getImageBounds().y + 
		    ((Integer)ge.getTextPosition().second).intValue());
	    D.o("Image y: "+ge.getImageBounds().y + 
		    "  Text y: "+y);
	    while (i.hasNext()) {
		String s = (String)i.next();
		// Centre each line.
		g.drawString(s, ge.getImageBounds().x + 
			ge.getImageBounds().width/2 -
			fm.stringWidth(s)/2, y);
		y += fh;
	    }
	}
    }


    /** Draws a link.
     *  @param 	the <code> Graphics </code> to draw on.
     *  @param gl	the link to be drawn.
     */
    public void drawLinks(Graphics2D g, GraphicViewLink gl) {
	GraphicViewObject from = (GraphicViewObject)gl.getFrom();
	GraphicViewObject to = (GraphicViewObject)gl.getTo();

	// images and lines are drawn by the system with a different bias,
	// however, we are using images for arrowheads...
	
	g.translate(-0.5,-0.5);
	Stroke s = g.getStroke();
	g.setStroke(gl.getStroke());
	g.drawLine(from.getCentre().x, from.getCentre().y,
			to.getCentre().x, to.getCentre().y);
	g.setStroke(s);
	g.translate(.5,.5);
	
	g.drawImage(gl.getImage(), gl.getImageBounds().x, 
		gl.getImageBounds().y, this);
	if (gl.getName() != null) {
	    Rectangle re = gl.getImageBounds();
	    g.drawString(gl.getName(), re.x+re.width, re.y);
	}
    }

    
    BufferedImage im3;
    Dimension im3s;
    Graphics2D im3g;

  
    /** draws all expanded, visible and not hidden objects. Called by Paint,
     * and Print to draw the elements and links, etc.. onto the canvas.
     *  @param g Graphics2D to draw onto
     */
    private void draw(Graphics2D g) {
	
	GraphicViewLink gl;
	GraphicViewElement ge;
	
	Iterator e;
	
	e = view.getExpandedElements().iterator();
	while(e.hasNext()) {
	    ge = (GraphicViewElement)e.next();
	    drawExpandedElements(g, ge);
	}   
	
	e = view.getVisibleLinks().iterator();
	while(e.hasNext()) {
	    
	    gl = (GraphicViewLink)e.next();
	    //D.o("From: " + gl.getFrom().getName() + "To: " + gl.getTo().getName());
	    if (! gl.isHidden() ) {
		drawLinks(g, gl);
	    }
	}

	e = view.getVisibleElements().iterator();
    
	while(e.hasNext()){
	    ge = (GraphicViewElement)e.next();
	    //D.o("Element: "+ge.getName());
	    if (! ge.isHidden() ) {
		drawElements(g,ge);
	    }
	}
	
	g.setColor(Color.red);
	g.setStroke(new BasicStroke((float).6));

	e = view.getSelectedLinks().iterator();
	while(e.hasNext()) {
	    gl = (GraphicViewLink)e.next();
	    if (! gl.isHidden() ) {
		g.drawRect(gl.getSelectableBounds().x,
			gl.getSelectableBounds().y,
			gl.getSelectableBounds().width-1,
			gl.getSelectableBounds().height-1);
	    }
	}
	
	
	e = view.getSelectedElements().iterator();
	while(e.hasNext()) {
	    ge = (GraphicViewElement)e.next();
	    if( !ge.isHidden() ) {
		g.drawRect(ge.getSelectableBounds().x,
			ge.getSelectableBounds().y,
			ge.getSelectableBounds().width-1,
			ge.getSelectableBounds().height-1);
	    }
	}
	
    }

    
    /** Exports the whole graphic view into an image file.
     *
     *  @param file the <code> File </code> where the graphic view should be exported
     */
    public synchronized void print(File file) {//String filename) {
	try {		    
	System.out.println("Start printing...");
	
	
	// "im3" is the image buffer. It is only (re)created when
	// necessary to save resource.
	Dimension d = new Dimension(view.getViewBounds().width,
	    view.getViewBounds().height);
	if ((im3 == null) || (d.width != im3s.width) || (d.height != im3s.height)) {
	    im3 = new BufferedImage(d.width, d.height,
		BufferedImage.TYPE_INT_BGR);
	    im3s = new Dimension(d.width, d.height);
	    im3g = (Graphics2D)im3.getGraphics();
	}

	
	im3g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
		RenderingHints.VALUE_ANTIALIAS_ON);
	im3g.setRenderingHint(RenderingHints.KEY_RENDERING,
		RenderingHints.VALUE_RENDER_QUALITY);
	im3g.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
		RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
	im3g.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
		RenderingHints.VALUE_INTERPOLATION_BICUBIC);
	
	im3g.setColor(getBackground());
	im3g.fillRect(0, 0, d.width, d.height);
	im3g.setFont(getFont());
	im3g.setColor(Color.black);
	im3g.setStroke(new BasicStroke((float)2.0));

	// Set display bounds for printing (save old ones)
	Rectangle r = view.getDisplayBounds();
	view.setDisplayBounds(view.getViewBounds());
	
	draw(im3g);
	view.setDisplayBounds(r);

	(new BMPWriter()).writeUncompressedBMP(im3, file, im3.getWidth(this),
					       im3.getHeight(this));
	System.out.println("Finish Printing...");
	}catch (Exception ex) {
	    ex.printStackTrace();
	    System.out.println(ex);
	}
    }

    /** Prompts the user for a filename, then exports the view to that file. */
    public void export() {
	
	try {
	    FileDialog exportDlg = new FileDialog(gframe, 
		    "Export File", FileDialog.SAVE);
	    exportDlg.setFile("*.png");
	    exportDlg.show();

	    String eFileName = exportDlg.getFile();
	    if(eFileName == null) return;
	    
	    File cd = new File(".");
	    File tempfile = File.createTempFile("tmp",".bmp",cd);
	    
	    print(tempfile); 
	    String cmd[] = new String[3];
	    cmd[0] = "..\\iview\\i_view32.exe";
	    cmd[1] = tempfile.getName();
	    cmd[2] = "/c="+eFileName;
	    D.o("converting to: " +cmd[2] );
	    Process p = Runtime.getRuntime().exec(cmd);
	    p.waitFor();
	    tempfile.delete();
	} catch (Exception ex) {
	    D.o(ex);
	}
    }
	

    
    Image im2;
    Dimension im2s;
    Graphics2D im2g;
    

    /** Draws visible elements and links on the screen.
     *  It only draws the visible elements and links to save resource.
     *
     *  @param g	the <code> Graphics </code> of the screen
     */
    public synchronized void paint(Graphics g) {

	//adjustScrollBar();

	System.out.println("Start painting...");

	// The "im2" is the image buffer for the paint function.
	// It is only (re)created when necessary.
	Dimension d = new Dimension(view.getDisplayBounds().width,
	    view.getDisplayBounds().height);
	if ((im2 == null) || (d.width != im2s.width) || (d.height != im2s.height)) {
	    im2 = createImage(d.width, d.height);
	    im2s = new Dimension(d.width, d.height);
	    im2g = (Graphics2D)im2.getGraphics();
	}
	
	im2g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
		RenderingHints.VALUE_ANTIALIAS_ON);
	
	// save displaybounds.
	Rectangle db = view.getDisplayBounds();

	// we must set the clip for our image to be that of the passed
	// Graphics object. interesting artifacts result if we are careless.
	Rectangle r = g.getClipBounds();
	im2g.setClip(r.x ,r.y , r.width, r.height);
	
	// clear display
	im2g.setColor(getBackground());
	im2g.fillRect(r.x,r.y,r.width,r.height);

	// Translate from scrollbars using *fullscreen displaybounds*.
	im2g.translate( -db.x, -db.y);
	
	// set font
	im2g.setFont(getFont());
	im2g.setColor(Color.black);
	im2g.setStroke(new BasicStroke((float)2.0));
	
	// set displaybounds to dirty area.
	view.setDisplayBounds(new Rectangle(r.x+db.x, r.y+db.y, r.width,
		    r.height));
	
	draw(im2g);
	    
	// Draws the image buffer onto the screen graphics.
	g.drawImage(im2, 0, 0, this);
	
	// restore displaybounds.
	view.setDisplayBounds(db);
	
	// Translate the co-ordinate back since im2 might be reused.
	im2g.translate(view.getDisplayBounds().x, view.getDisplayBounds().y);

	paintChildren(g);
	//view.setViewClean();
	
	System.out.println("Finish Painting...");
    }


    /** Repaints the screen.
     *
     *  @param g	the <code> Graphics </code> of the screen.
     */
    public synchronized void update(Graphics g) {
	paint(g); 
    }
   

    public boolean isTabbed;
    private TabbedView tv;

    /** Displays the rename box. 
     *
     *  @param gve	the graphic view object which we wish to rename 
     */
    /*package*/ void openRenameBox(GraphicViewObject gve) {
	int sizex = gve.getSelectableBounds().width/2;
	int sizey = gve.getSelectableBounds().height/2;

	JTextField rb = new JTextField(15);
	add(rb);
	rb.setBounds(gve.getSelectableBounds().x - view.getDisplayBounds().x, 
		gve.getSelectableBounds().y - view.getDisplayBounds().y 
		+ sizey - 12, 2*sizex, 24);
	rb.addKeyListener(new RenameBoxKeyAdapter(gve, rb, this));
	rb.addFocusListener(new RenameBoxFocusAdapter(gve,rb,this));

	rb.setText(gve.getName());
	rb.selectAll();
	rb.repaint();
	rb.requestFocus();
	D.o("creating Jtextfield at:" + rb.getBounds());
	D.o("renamebox created.");
    }
   

    /** Closes the rename box. 
     *
     *  @param tf 	the text Field in the rename in rename box. It needs
     *			to be removed so we don't see it anymore.
     */
    private void closeRenameBox(JTextField tf) {
	D.o("Closing rename box at " + tf.getBounds());
	remove(tf);
	repaint();
    }


    /** Displays the resize box. 
     *  @param gve	the graphic view element we wish to resize
     */
    /*package*/ void openResizeBox(GraphicViewObject gvo) {
	int sizex = gvo.getSelectableBounds().width/2;
	int sizey = gvo.getSelectableBounds().height/2;

	JTextField tf = new JTextField(15);
	tf.setBounds(gvo.getSelectableBounds().x - view.getDisplayBounds().x, 
		gvo.getSelectableBounds().y - view.getDisplayBounds().y + sizey - 12,
		2*sizex, 24);
	tf.addKeyListener(new ResizeBoxKeyAdapter(gvo, tf, this));
	tf.addFocusListener(new ResizeBoxFocusAdapter(gvo,tf,this));

	tf.setText(""+gvo.getScale());
	add(tf);
	tf.selectAll();
	tf.repaint();
	tf.requestFocus();
	D.o("resizebox created.");
    }
   

    /** Closes the resize box. 
     *  @param tf	the text field in the resize box.  It needs to be
     *			removed so we don't see it anymore.
     */
    private void closeResizeBox(JTextField tf) {
	remove(tf);
	repaint();
    }
    

    /** Saves the clicked point. 
     *  @param p	the clicked point
     */
    private void setClickedP(Point p) {
	clickedp = p;
    }
    
       
    /** Recovers the clicked point. */
    private Point getClickedP() {
	return clickedp;
    }

    /** Sets the input mode for the GVC */
    private void setMode(int mode, String message) {
	switch (mode) {
	    case NORMAL_MODE:
		setNormalMode(message);
		break;
	    case GET_PARAMETER:
		setGPMode(message);
		break;
	}
    }

    /** Go into NORMAL_MODE */
    private void setNormalMode(String message) {
	mode = NORMAL_MODE;
	gframe.getStatusBar().setMessage(message);
	setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	view.deselectAll();
    }

    /** Go into GET_PARAMETER mode */
    private void setGPMode(String message) {	
	mode = GET_PARAMETER;
	gframe.getStatusBar().setMessage(message);
	setCursor(new Cursor(Cursor.HAND_CURSOR));
    }
	

    /** Returns our mode. */
    private int getMode() {
	return mode;
    }

    /** Returns our GVCParameterCollector */
    private GVCParameterCollector getParameterCollector() {
	return pcollector;
    }
    
    /** Sets the title of our frame. */
    /*package*/ void setTitle(String s) {
	gframe.setTitle(s);
    }

    // Save will be ready to die soon. 
    /** Saves the view being displayed on this canvas. */
    ///*
    public void save() {
	String sFileName = null;
	if (true) {
	    FileDialog saveDlg = new FileDialog(gframe, "Save File",
		FileDialog.SAVE);
	    //saveDlg.setFilenameFilter(new TelosFileFilter());
	    saveDlg.setFile("*.tel");
	    //saveDlg.setFile(view.getFilename());
	    saveDlg.show();

	    sFileName = saveDlg.getFile();
	    if(sFileName == null) return;
	    m_sCurrFile = sFileName;
	    m_sCurrDir = saveDlg.getDirectory();
	    setTitle(m_sCurrDir + m_sCurrFile);
	}
	
	// serialize all the views to the model.
	// 
	// save the  the Model:
	
	try {
	    view.getModel().getModelManager().
		    saveModelTo(view.getModel(), sFileName,
			    m_sCurrDir);
	} catch (Exception ex) {
	    D.o(ex);
	}
    }
    //*/

    /** Throws up a save dialogue and returns a collection whose first
     *  member is the file name and whose second member is the directory name
     *  (both of which a Strings).  Returns a null vector if no file was
     *  selected by the user.
     *	Side effect:  Sets m_sCurrFile and m_sCurrDir to names supplied by
     *  user.  Also sets the title of our frame to the pathname of the file.
     */
    public Collection getSaveInformation() {
	String sFileName;
	Vector v = null;
	
	FileDialog saveDlg = new FileDialog(gframe, "Save File",
		FileDialog.SAVE);
	//saveDlg.setFilenameFilter(new TelosFileFilter());
	saveDlg.setFile("*.tel");
	//saveDlg.setFile(view.getFilename());
	saveDlg.show();

	sFileName = saveDlg.getFile();
	if(sFileName == null) { 
	    return v;
	} // else
	m_sCurrFile = sFileName;
	m_sCurrDir = saveDlg.getDirectory();
	setTitle(m_sCurrDir + m_sCurrFile);

	v = new Vector(2);
	v.add(m_sCurrFile);
	v.add(m_sCurrDir);
	return v;
    }
    
    /** The keyboard listener for <code> GraphicViewCanvas </code>. */
    class GVCKeyAdapter extends KeyAdapter {

	/** Processes keyboard input.
	 *
	 *  @param e	the keyboard event to be processed
	 */
	public void keyPressed (KeyEvent e) {
	    char c = e.getKeyChar();
	    
	    switch (c) {
		case KeyEvent.VK_DELETE:
		    view.deleteElements();
		    view.deleteLinks();
		    repaint();
		    if (isTabbed) tv.refresh();
		    break;

		case KeyEvent.VK_ESCAPE:
		    if (getMode() == GET_PARAMETER) {
			// User wants out of the current plugin method.
			getParameterCollector().cancelMethod();
		    } else  { //normal mode
			view.deselectAll();
			repaint();
		    }

	    }
	}
    }


    /** The mouse event listener for <code> GraphicViewCanvas </code>. */
    class GVCMouseAdapter extends MouseAdapter {

	private int initialmode;
	
	/** Processes mousePressed event. This is the mouse button coming down,
	 *  on which we select things
	 *
	 *  @param e	the mouse event to be processed
	 */
	public void mousePressed( MouseEvent me ) {
	    System.out.println("mousepressed");
	    requestFocus();

	    Point p = view.toView(me.getPoint());
	    setClickedP (p);

	    initialmode = getMode();

	    if (initialmode == GET_PARAMETER) {
		getParameterCollector().processMousePressed(me);	
	    } else {
		
		GraphicViewElement ge = view.findElement(p);	    
		if(ge == null) pressed = 0;
		else if (view.getSelectedElements().contains(ge)) pressed = 2;
		else pressed = 1;
		if ((me.getModifiers() &InputEvent.BUTTON3_MASK) == 0 ) {
		    if((me.getModifiers() & InputEvent.CTRL_MASK) == 0) {
			select(p);
		    }
		}
	    }
	}
	

	/** Processes mouseClicked event.
	 *  @param me	the mouse event to be processed
	 *  I should throw actions!!
	 */
	public void mouseClicked( MouseEvent me ) {
	    if (initialmode == GET_PARAMETER) {
		// Ignore this event.
	    } else {
	    
		requestFocus();
		
		System.out.println("mouseclicked: " + me.getClickCount());
		Point p = view.toView(me.getPoint());
		setClickedP(p);
		
		if((me.getModifiers() & InputEvent.BUTTON3_MASK) == 0) {
		    // regular button has been hit.
		    if (me.getClickCount() == 2) { 
			// on double click:
			GraphicViewElement gve = view.findElement(p);
			if (gve != null) {
			    try {
			    if (view.getExpandedElements().contains(gve) ) {
				view.contractElements();
			    } else {	
				if(gve.isExpandable()) {
				    view.expandElements();
				}
			    }
			    } catch (Exception e) {
				D.o(e);
			    }
			    repaint();
			}
			//if (view.isViewDirty()) adjustScrollBar();
			//repaint();
		    } else {
			// on single (triple. quadruple?) click
		      
			if((me.getModifiers() & InputEvent.CTRL_MASK) != 0) {
			    toggleSelect(p);
			} else {
			    select(p);
			}
			repaint();
		    }
		} else {
		    // Alternate button has been hit. Show the context menu.
		    // if (popupmenu == null) {
			popupmenu = setPopupMenu(p);
		    // }
		    popupmenu.show(me.getComponent(), me.getX(), me.getY());
		    
		}
	    }
	}
    }


    /** The mouse motion listener for <code> GraphicViewCanvas </code>. */
    class GVCMouseMotionAdapter extends MouseMotionAdapter {

	/** Processes mouseDragged event.
	 *
	 *  @param me	the mouse event to be processed
	 */
	public void mouseDragged( MouseEvent me ) {

	    if (getMode() == GET_PARAMETER) {
		// Ignore drags
	    } else if ((me.getModifiers() & InputEvent.BUTTON3_MASK) == 0) {
		System.out.println("mousedragged");
		
		Point p = view.toView(me.getPoint());
		view.moveElements(p.x - getClickedP().x, p.y -
			getClickedP().y);
		    
		setClickedP(p);
		Rectangle dr = view.getDirtyRect();
		Rectangle db = view.getDisplayBounds();
		if (dr !=null) {
		    repaint(dr.x-db.x, dr.y-db.y, dr.width, dr.height);
		}
		adjustScrollBar();		
	    }
	}
    }
	   

    /** The action listener used for Menu(bar) and Toolbar buttons. */
    class GVCActionAdapter implements ActionListener {
	private OMEMethod method;
	private GraphicViewCanvas gvc;
	private Point p = null;
	private JToggleButton togglebutton = null;

	/**  Create a new GVCActionAdapter.
	 *   @param method	The associated method.
	 *   @param p		The Point associated with the firing of this
	 *			action (if any).
	 *   @param gvc		Our canvas
	 */
	GVCActionAdapter(OMEMethod method, Point p, 
		GraphicViewCanvas gvc) {
	    this.gvc = gvc;
	    this.method = method;
	    this.p = p;
	}
	
	/**  Create a new GVCACtionAdapter.
	 *   @param method	The associated method.
	 *   @param gvc		Our canvas
	 */
	GVCActionAdapter(OMEMethod method, 
		GraphicViewCanvas gvc) {
	    this.gvc = gvc;
	    this.method = method;
	}
	
	/**  Create a new GVCACtionAdapter.
	 *   @param method	    The associated method.
	 *   @param togglebutton    The button to untoggle when done.
	 *   @param gvc		    Our canvas
	 */
	GVCActionAdapter(OMEMethod method, JToggleButton togglebutton, 
		GraphicViewCanvas gvc) {
	    this.gvc = gvc;
	    this.method = method;
	    this.togglebutton = togglebutton;
	}

	/** Process the menu and tool bar events.
	 *  @param e	the action event to be processed
	 */
	public void actionPerformed(ActionEvent e) {
	    // If we have a toggle button, and the button is currently NOT
	    // selected, then the user has UNSELECTED it, and we should
	    // cancel the method.
	    if (togglebutton != null && !togglebutton.isSelected()) {
		getParameterCollector().cancelMethod();
	    } else {
		OMEContext c = view.generateContext(p);
		gvc.getParameterCollector().startCollecting(method, c,
			togglebutton);
	    }
	}	    
    }

    /** The component listener for the <code> GraphicViewCanvas </code>. */
    class GVCComponentAdapter extends ComponentAdapter {

	/** Processes the resized event.
	 *
	 *  @param e	the component event to be processed
	 */
	public void componentResized( ComponentEvent e) {
	    view.getDisplayBounds().width = getSize().width;
	    view.getDisplayBounds().height = getSize().height;
	   
	    // Adjusts the value in GraphicView.
	    view.getDisplayBounds().x =
		Math.max(view.getViewBounds().x,
		    Math.min(
		    view.getViewBounds().width-view.getDisplayBounds().width+view.getViewBounds().x,
		    view.getDisplayBounds().x));
	    
	    view.getDisplayBounds().y =
		Math.max(view.getViewBounds().y,
		    Math.min(
		    view.getViewBounds().height-view.getDisplayBounds().height+view.getViewBounds().y,
		    view.getDisplayBounds().y));
	   
	    adjustScrollBar();
	    // The lists of visible elements and links are changed.
	    view.refreshVisibleLists();
	    super.componentResized( e );
	}
    }

    /** The key listener for the rename box. */
    class RenameBoxKeyAdapter extends KeyAdapter {

	private GraphicViewObject gve;
	private JTextField tf;
	private GraphicViewCanvas gvc;
	
	/** Creates a new RenameBoxKeyAdapter.
         *  @param gve	The GraphicViewElement being resized.
         *  @param tf 	The textfield in the Resize box.
         *  @param gvc 	The GraphicViewCanvas to paint on.
         */
	public RenameBoxKeyAdapter(GraphicViewObject gve, JTextField tf,
			GraphicViewCanvas gvc) {
	    super();
	    this.gve=gve;
	    this.tf=tf;
	    this.gvc=gvc;
	}	    	    
	    
	/** Processes keyPressed event.
	 *
	 *  @param e	the keyboard event to be processed
	 */
	public void keyPressed (KeyEvent e) {
	    char c = e.getKeyChar();
	    if (c == KeyEvent.VK_ENTER) {
		//closeRenameBox();
		try {
		    gve.setName(tf.getText());
		} catch (Exception ex) {
		    D.o("gve.setName exception");
		}
		closeRenameBox(tf);
		//gvc.remove(tf);
		//gvc.repaint();
	    }
	}
    }
    
    
    /** JOE please comment */
    class RenameBoxFocusAdapter extends FocusAdapter {

	private GraphicViewObject gve;
	private JTextField tf;
	private GraphicViewCanvas gvc;
	 
      /** Create a new RenameBoxFocusAdapter.
       *  @param gve	The GraphicViewElement being resized.
       *  @param tf 	The textfield in the Resize box.
       *  @param gvc 	The GraphicViewCanvas to paint on.
       */
	public RenameBoxFocusAdapter(GraphicViewObject gve, JTextField tf,
			GraphicViewCanvas gvc) {
	    super();
	    this.gve=gve;
	    this.tf=tf;	    
	    this.gvc=gvc;
	    D.o("Created RenameBoxFocusAdapter");
	}


      /** Process the focusLost event.
       *  @param e	The focus event we are processing.
       */
	public void focusLost(FocusEvent e) {
	    if ((e.getID() == FocusEvent.FOCUS_LOST) && !e.isTemporary()) {
		try {
		    gve.setName(tf.getText());
		} catch (Exception ex) {
		    D.o("gve.setName exception");
		}
		//gvc.remove(tf);
		//gvc.repaint();
		D.o("focusbox closed.");
		closeRenameBox(tf);
	    }
	}
    }
		

    /** JOE please comment */
    class ResizeBoxKeyAdapter extends KeyAdapter {
	
      private GraphicViewObject gve;
	private JTextField tf;
	private GraphicViewCanvas gvc;

      /** Creates a new ResizeBoxKeyAdapter
       *  @param gve	The GraphicViewElement being resized.
       *  @param tf 	The textfield in the Resize box.
       *  @param gvc 	The GraphicViewCanvas to paint on.
       */
	public ResizeBoxKeyAdapter(GraphicViewObject gve, JTextField tf,
		GraphicViewCanvas gvc) {
	    super ();
	    this.gve=gve;
	    this.tf=tf;	    
	    this.gvc=gvc;
	    D.o("Resize box key adapter created.");
	}


      /** Handle a key pressed event.
       * 
       *  @param e	the key-pressed event to be handled
       */
	public void keyPressed(KeyEvent e) {
	    char c = e.getKeyChar();
	    if(c == KeyEvent.VK_ENTER) {
		try {
		    Double d = new Double(tf.getText());
		    gve.setScale(d.doubleValue());
		    gvc.remove(tf);
		} catch (NumberFormatException nfe) {
		    //beep!
		    tf.setText(""+gve.getScale());
		    tf.selectAll();
		} finally {
		    gvc.repaint();
		}
	    }
	    D.o("voodoo");
	}
    }

    
    /** JOE Please comment */
    class ResizeBoxFocusAdapter extends FocusAdapter {
	private GraphicViewObject gve;
	private JTextField tf;
	private GraphicViewCanvas gvc;
	 
      /** Creates a new ResizeBoxFocusAdapter.
       *  @param gve	The GraphicViewElement being resized.
       *  @param tf 	The textfield in the Resize box.
       *  @param gvc 	The GraphicViewCanvas to paint on.
       */
	public ResizeBoxFocusAdapter(GraphicViewObject gve, JTextField tf,
			GraphicViewCanvas gvc) {
	    super();
	    this.gve=gve;
	    this.tf=tf;	    
	    this.gvc=gvc;
	    D.o("Created ResizeBoxFocusAdapter");
	}

	public void focusLost(FocusEvent e) {
	    if ((e.getID() == FocusEvent.FOCUS_LOST) && !e.isTemporary()) {
		try {
		    Double d = new Double(tf.getText());
		    gve.setScale(d.doubleValue());		    
		} catch (NumberFormatException nfe) { // we throw it away.
		} finally {
		    gvc.remove(tf);
		    gvc.repaint();
		}
	    }
	}
    }
}
